
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define LEN	(1024*2+1)

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static void	Abort(path,msg)
char		*path,*msg;
{
	(void)fputs(path,stderr); (void)fputs(" : ",stderr); Error(msg);
}

#define GET1(file)	fgetc(file)

static int	Get2(file)
FILE		*file;
{
	int	lo,hi;

	return ((lo=GET1(file))==EOF ||
		(hi=GET1(file))==EOF)?-1:(hi<<8)|lo;
}

typedef struct List {
		char		*org,*new;
		long		offset;
		struct List	*next;
	} List;

static char	*ueof="unexpected end of file.",
		*mae="memory allocation error.";

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

static List	*CheckHiPic(org,new,next,Nh,Nv)
char		*org,*new;
List		*next;
int		*Nh,*Nv;
{
	FILE	*file;
	int	cl,h,v;
	long	offset;
	List	*list;

	if ((file=fopen(org,"rb"))==NULL) Abort(org,"file not found.");

	if (GET1(file)!='I' ||
	    GET1(file)!='M') Abort(org,"bad magic number.");

	if ((cl=Get2(file))<0) Abort(org,"bad comment length.");

	if ((h=Get2(file))<=0 ||
	    (v=Get2(file))<=0) Abort(org,"bad image size.");

	if (next==NULL) {
	    *Nh=h;
	    *Nv=v;
	}
	else
	    if (*Nh!=h ||
		*Nv!=v) Abort(org,"image size not match.");

	(void)fseek(file,4L,SEEK_CUR);

	if (Get2(file)!=2) Abort(org,"bad image type.");

	(void)fseek(file,(long)(50+cl),SEEK_CUR);

	if ((offset=ftell(file))<0) Abort(org,ueof);

	(void)fclose(file);

	if ((list=ALLOC(List,1))==NULL ||
	    (list->org=ALLOC(char,strlen(org)+1))==NULL ||
	    (list->new=ALLOC(char,strlen(new)+1))==NULL) Error(mae);

	(void)strcpy(list->org,org);
	(void)strcpy(list->new,new);
	list->offset=offset;
	list->next=next;
	return list;
}

#define PUT1(data,file)	(void)fputc(data,file)
#define PUT2(data,file)	PUT1(data&0xff,file); PUT1((data>>8)&0xff,file)

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	base,step,H,V,r1,r2,norm,h0,v0,dh,dv,hr,hz,vr,vz;
	int	null,Nh,Nv,v,R,Z,Nr,Nz,h,pv,idx,z,r;
	char	line[LEN],org[LEN],new[LEN];
	List	*next,*list=NULL;
	short	**PV;
	FILE	*file;

	if (argc!=4 && argc!=5 && argc!=6)
	    Error("usage : hp_irac base step null {org.img {new.img}}");

	base=atof(argv[1]);
	step=atof(argv[2]);

	if ((null=atoi(argv[3]))<0 || null>=65536) Error("bad null value.");

	if (argc==4) {
	    while (fgets(line,LEN,stdin)!=NULL) {
		if (strchr(line,'\n')==NULL) Error("too long path name(s).");

		switch(sscanf(line,"%s %s",org,new)) {
		    case 1 : list=CheckHiPic(org,org,list,&Nh,&Nv); break;
		    case 2 : list=CheckHiPic(org,new,list,&Nh,&Nv); break;
		}
	    }
	    if (list==NULL) Error("no path name.");
	}
	else
	    list=CheckHiPic(argv[4],argv[argc-1],list,&Nh,&Nv);

	(void)printf("%d\t%d\n",Nh,Nv);

	if ((PV=ALLOC(short *,Nv))==NULL) Error(mae);

	for (v=0; v<Nv; v++)
	    if ((PV[v]=ALLOC(short,Nh))==NULL) Error(mae);

	H=(double)(Nh-1);
	V=(double)(Nv-1);

	if (step>=0.0) {
	    r1=base+step*V; r2=base-H;
	}
	else {
	    r1=base; r2=base+step*V-H;
	}
	if (r1*r2>0.0) Error("bad rotation axis.");

	norm=sqrt(step*step+1.0);

	r1=fabs(r1);
	r2=fabs(r2); R=(int)(((r1<=r2)?r1:r2)/norm);

	h0=base+step*(v0=0.0);
	dh=base+step*(dv=V  );
	if (step>=0.0) {
	    if (h0<0.0) v0=((h0=0.0)-base)/step;
	    if (dh>H  ) dv=((dh=H  )-base)/step;
	}
	else {
	    if (h0>H  ) v0=((h0=H  )-base)/step;
	    if (dh<0.0) dv=((dh=0.0)-base)/step;
	}
	dh-=h0;
	dv-=v0; Z=(int)sqrt(dh*dh+dv*dv);

	Nr=R*2+1;
	Nz=Z  +1; vr=(-(hz=step*(hr=vz=1.0/norm)));

	(void)printf("%d\t%d\n",Nr,Nz);

	do {
	    if ((file=fopen(list->org,"rb"))==NULL ||
		fseek(file,list->offset,SEEK_SET))
		Abort(list->org,"file not open.");

	    for (v=0; v<Nv; v++)
	    for (h=0; h<Nh; h++) {
		if ((pv=Get2(file))<0) Abort(list->org,ueof);

		PV[v][h]=(short)pv;
	    }
	    (void)fclose(file);

	    if ((file=fopen(list->new,"wb"))==NULL)
		Abort(list->new,"can not create new file.");

	    (void)fputs("IM",file);			/* magic number */

	    PUT2(0,file);				/* comment length */
	    PUT2(Nr,file);				/* image width */
	    PUT2(Nz,file);				/* image height */

	    for (idx=0; idx<4; idx++) PUT1(0,file);	/* image offset */

	    PUT2(2,file);				/* image type */

	    for (idx=0; idx<50; idx++) PUT1(0,file);	/* reserved data */

	    for (z=  0 ; z<=Z; z++)
	    for (r=(-R); r<=R; r++) {
#ifdef	BLI
		h=(int)floor(dh=h0+hr*(double)r+hz*(double)z); dh-=(double)h;
		v=(int)floor(dv=v0+vr*(double)r+vz*(double)z); dv-=(double)v;
{
		int	h1=h+1,h2=(h>=0 && h<Nh),h3=(h1>=0 && h1<Nh),
			v1=v+1,v2=(v>=0 && v<Nv),v3=(v1>=0 && v1<Nv);
		double	pv00=(h2 && v2)?(double)PV[v ][h ]:(double)null,
			pv01=(h3 && v2)?(double)PV[v ][h1]:(double)null,
			pv10=(h2 && v3)?(double)PV[v1][h ]:(double)null,
			pv11=(h3 && v3)?(double)PV[v1][h1]:(double)null;

		pv=(int)(pv00+(pv01-pv00)*dh
			     +(pv10-pv00)*dv+(pv11-pv10-pv01+pv00)*dh*dv+0.5);
}
#else
#define ROUND(data)	(int)floor(data+0.5)

		pv=((h=ROUND(h0+hr*(double)r+hz*(double)z))>=0 && h<Nh &&
		    (v=ROUND(v0+vr*(double)r+vz*(double)z))>=0 && v<Nv)
		   ?PV[v][h]:null;
#endif
		PUT2(pv,file);
	    }
	    (void)fclose(file);

	    next=list->next; (void)free(list);
	} while ((list=next)!=NULL);

	return 0;
}
