
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

static void	Error(head,tail)
char		*head,*tail;
{
	if (head!=NULL) (void)fprintf(stderr,"%s : ",head);

	(void)fputs(tail,stderr); (void)fputc('\n',stderr); exit(1);
}

static int	FGetW(file)
FILE		*file;
{
	int	lo,hi;

	return ((lo=fgetc(file))==EOF || (hi=fgetc(file))==EOF)?-1:(hi<<8)|lo;
}

static int	Nx,Ny,first=1;

static FILE	*Open(path)
char		*path;
{
	FILE	*file;
	int	cl;

	if ((file=fopen(path,"rb"))==NULL) Error(path,"file not found.");

	if (fgetc(file)!='I' ||
	    fgetc(file)!='M') Error(path,"bad magic number.");

	if ((cl=FGetW(file))<0) Error(path,"bad comment length.");

	if (first) {
	    if ((Nx=FGetW(file))<=0 ||
		(Ny=FGetW(file))<=0) Error(path,"bad image size.");

	    first=0;
	}
	else
	    if (Nx!=FGetW(file) ||
		Ny!=FGetW(file)) Error(path,"image size not match.");

	(void)fseek(file,4L,SEEK_CUR);

	if (FGetW(file)!=2) Error(path,"bad image type.");

	(void)fseek(file,(long)(50+cl),SEEK_CUR);

	return file;
}

static int	X1,Y1,X2,Y2,nx,ny;
static long	Oy,Ox1,Ox2;

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(size_t)(noe))

static double	**Read(Pd,Fd,Pi,Fi,Pe,Fe,P)
char		*Pd,*Pi,*Pe,*P;
FILE		*Fd,*Fi,*Fe;
{
	double	**p;
	char	*mae="memory allocation error.",
		*ueof="unexpected end of file.";
	int	y,x,d,i,e,No=0;

	if ((p=ALLOC(double *,ny))==NULL) Error(P,mae);

	p-=Y1;

	(void)fseek(Fd,Oy,SEEK_CUR);
	(void)fseek(Fi,Oy,SEEK_CUR);
	(void)fseek(Fe,Oy,SEEK_CUR);

	for (y=Y1; y<=Y2; y++) {
	    if ((p[y]=ALLOC(double,nx))==NULL) Error(P,mae);

	    p[y]-=X1;

	    (void)fseek(Fd,Ox1,SEEK_CUR);
	    (void)fseek(Fi,Ox1,SEEK_CUR);
	    (void)fseek(Fe,Ox1,SEEK_CUR);

	    for (x=X1; x<=X2; x++) {
		if ((d=FGetW(Fd))<0) Error(Pd,ueof);
		if ((i=FGetW(Fi))<0) Error(Pi,ueof);
		if ((e=FGetW(Fe))<0) Error(Pe,ueof);

		if (i<=d || e<=d)
		    p[y][x]=HUGE_VAL;
		else {
		    p[y][x]=log((double)(i-d)/(double)(e-d)); ++No;
		}
	    }
	    (void)fseek(Fd,Ox2,SEEK_CUR);
	    (void)fseek(Fi,Ox2,SEEK_CUR);
	    (void)fseek(Fe,Ox2,SEEK_CUR);
	}
	if (No==0) Error(P,"no normal data.");

	(void)fclose(Fe);
	(void)fclose(Fi);
	(void)fclose(Fd); return p;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*Fd000,*Fi000,*Fe000,*Fd180,*Fi180,*Fe180;
	char	*bia="bad image area.";
	double	B0,dB,S0,dS,**p000,**p180,S,C0,C1,C2,B,C3,C4,SD,X0,Y0,dp;
	int	nB,nS,b,s,No,y,x,X,Y;

	if (argc!=13 && argc!=15 && argc!=17)
	    Error("usage","ira_rmsd D000 I000 E000 D180 I180 E180\n"
		  "                 {y1 y2}|{x1 y1 x2 y2}\n"
		  "                 B0 dB nB S0 dS nS");

	Fd000=Open(argv[1]); Fi000=Open(argv[2]); Fe000=Open(argv[3]);
	Fd180=Open(argv[4]); Fi180=Open(argv[5]); Fe180=Open(argv[6]);

	switch(argc) {
	    case 13 :
		X1=Y1=0; X2=Nx-1; Y2=Ny-1; break;
	    case 15 :
		if ((Y1=atoi(argv[7]))<0 || (Y2=atoi(argv[8]))>=Ny || Y1>Y2)
		    Error((char *)NULL,bia);

		X1=0; X2=Nx-1; break;
	    case 17 :
		if ((X1=atoi(argv[7]))<0 || (X2=atoi(argv[ 9]))>=Nx || X1>X2 ||
		    (Y1=atoi(argv[8]))<0 || (Y2=atoi(argv[10]))>=Ny || Y1>Y2)
		    Error((char *)NULL,bia);

		break;
	}
	nx=X2-X1+1; ny=Y2-Y1+1; Oy=Y1*Nx*2; Ox1=X1*2; Ox2=(Nx-1-X2)*2;

	p000=Read(argv[1],Fd000,
		  argv[2],Fi000,
		  argv[3],Fe000,"projection from 000 degree");
	p180=Read(argv[4],Fd180,
		  argv[5],Fi180,
		  argv[6],Fe180,"projection from 180 degree");

	B0=atof(argv[argc-6]); dB=atof(argv[argc-5]); nB=atoi(argv[argc-4]);
	S0=atof(argv[argc-3]); dS=atof(argv[argc-2]); nS=atoi(argv[argc-1]);

	for (s=0; s<nS; s++) {
	    S=S0+dS*(double)s; C0=S*S+1.0; C1=(S*S-1.0)/C0; C2=2.0*S/C0;
	    for (b=0; b<nB; b++) {
		B=B0+dB*(double)b; C3=2.0*B/C0+0.5; C4=B*C2-0.5; SD=0.0; No=0;
		for (y=Y1; y<=Y2; y++) {
		    X0=C2*(double)y+C3; Y0=C1*(double)y+C4;
		    for (x=X1; x<=X2; x++)
			if (p000[y][x]!=HUGE_VAL &&
			    (X=(int)(C1*(double)x+X0))>=X1 && X<=X2 &&
			    (Y=(int)(C2*(double)x-Y0))>=Y1 && Y<=Y2 &&
			    p180[Y][X]!=HUGE_VAL) {
			    dp=p000[y][x]-p180[Y][X]; SD+=(dp*dp); ++No;
			}
		}
		if (No!=0)
		(void)printf("%lf\t%lf\t%18.16lf\n",B,S,sqrt(SD/(double)No));
	    }
	}
	return 0;
}
