
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

#define LEN	(1024*2+1)

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

typedef struct List {
		char		*org,*new;
		struct List	*next;
	} List;

static char	*mae="memory allocation error.";

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

static List	*CheckTIFF(org,new,next,Nh,Nv,BPS)
char		*org,*new;
List		*next;
int		*Nh,*Nv,*BPS;
{
	int	h,v,bps;
	List	*list;

	ReadImageFile(org,&h,&v,&bps,NULL,NULL);

	if (next==NULL) {
	    *Nh=h;
	    *Nv=v; *BPS=bps;
	}
	else {
	    if (*Nh!=h ||
		*Nv!=v) {
		(void)fputs(org,stderr); Error(" : image size not match.");
	    }
	    if (*BPS>bps) *BPS=bps;
	}
	if ((list=ALLOC(List,1))==NULL ||
	    (list->org=ALLOC(char,strlen(org)+1))==NULL ||
	    (list->new=ALLOC(char,strlen(new)+1))==NULL) Error(mae);

	(void)strcpy(list->org,org);
	(void)strcpy(list->new,new);
	list->next=next;
	return list;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	base,step,H,V,r1,r2,norm,h0,v0,dh,dv,hr,hz,vr,vz;
	int	Nh,Nv,BPS,null,R,Z,Nr,Nz,z,r,h,v;
	char	line[LEN],org[LEN],new[LEN];
	List	*next,*list=NULL;
	Cell	**cell,**PV,*c;

	if (argc!=4 && argc!=5 && argc!=6)
	    Error("usage : tiff_irac base step null {org.tif {new.tif}}");

	base=atof(argv[1]);
	step=atof(argv[2]);

	if (argc==4) {
	    while (fgets(line,LEN,stdin)!=NULL) {
		if (strchr(line,'\n')==NULL) Error("too long path name(s).");

		switch(sscanf(line,"%s %s",org,new)) {
		    case 1 : list=CheckTIFF(org,org,list,&Nh,&Nv,&BPS); break;
		    case 2 : list=CheckTIFF(org,new,list,&Nh,&Nv,&BPS); break;
		}
	    }
	    if (list==NULL) Error("no path name.");
	}
	else
	    list=CheckTIFF(argv[4],argv[argc-1],list,&Nh,&Nv,&BPS);

	if ((null=atoi(argv[3]))<0 || null>=1<<BPS) Error("bad null value.");

	(void)printf("%d\t%d\n",Nh,Nv);

	H=(double)(Nh-1);
	V=(double)(Nv-1);

	if (step>=0.0) {
	    r1=base+step*V; r2=base-H;
	}
	else {
	    r1=base; r2=base+step*V-H;
	}
	if (r1*r2>0.0) Error("bad rotation axis.");

	norm=sqrt(step*step+1.0);

	r1=fabs(r1);
	r2=fabs(r2); R=(int)(((r1<=r2)?r1:r2)/norm);

	h0=base+step*(v0=0.0);
	dh=base+step*(dv=V  );
	if (step>=0.0) {
	    if (h0<0.0) v0=((h0=0.0)-base)/step;
	    if (dh>H  ) dv=((dh=H  )-base)/step;
	}
	else {
	    if (h0>H  ) v0=((h0=H  )-base)/step;
	    if (dh<0.0) dv=((dh=0.0)-base)/step;
	}
	dh-=h0;
	dv-=v0; Z=(int)sqrt(dh*dh+dv*dv);

	Nr=R*2+1;
	Nz=Z  +1; vr=(-(hz=step*(hr=vz=1.0/norm)));

	(void)printf("%d\t%d\n",Nr,Nz);

	if ((cell=ALLOC(Cell *,Nz))==NULL) Error(mae);

	for (z=0; z<Nz; z++)
	    if ((cell[z]=ALLOC(Cell,Nr))==NULL) Error(mae);

	do {
	    ReadImageFile(list->org,NULL,NULL,&BPS,&PV,NULL);

	    for (	    z=  0 ; z<=Z; z++	  )
	    for (c=cell[z], r=(-R); r<=R; r++, c++)
#ifdef	BLI
{
		int	h1=(h=(int)floor(dh=h0+hr*(double)r+hz*(double)z))+1,
			v1=(v=(int)floor(dv=v0+vr*(double)r+vz*(double)z))+1,
			h2=(h>=0 && h<Nh),
			v2=(v>=0 && v<Nv),
			h3=(h1>=0 && h1<Nh),
			v3=(v1>=0 && v1<Nv);
		double	pv00=(h2 && v2)?(double)PV[v ][h ]:(double)null,
			pv01=(h3 && v2)?(double)PV[v ][h1]:(double)null,
			pv10=(h2 && v3)?(double)PV[v1][h ]:(double)null,
			pv11=(h3 && v3)?(double)PV[v1][h1]:(double)null,
			h4=dh-(double)h,
			v4=dv-(double)v;

		*c=(Cell)(pv00+(pv01-pv00)*h4
			      +(pv10-pv00)*v4+(pv11-pv10-pv01+pv00)*h4*v4+0.5);
}
#else
#define ROUND(data)	(int)floor(data+0.5)

		*c=((h=ROUND(h0+hr*(double)r+hz*(double)z))>=0 && h<Nh &&
		    (v=ROUND(v0+vr*(double)r+vz*(double)z))>=0 && v<Nv)
		   ?PV[v][h]:(Cell)null;
#endif
	    StoreImageFile(list->new,Nr,Nz,BPS,cell,NULL);

	    for (v=Nv-1; v>=0; v--) free(PV[v]); free(PV);

	    next=list->next; (void)free(list);
	} while ((list=next)!=NULL);

	return 0;
}
